from dios.lib import *
from dios.options import *
from dios.itypes import *
from dios.errors import *
import pandas as pd
import numpy as np
import operator as op
from functools import partialmethod

from functools import wraps
from collections import OrderedDict
from pandas._libs.lib import to_object_array, is_bool_array
from pandas.core.common import is_bool_indexer
from pandas.core.dtypes.common import (
    is_list_like,
    is_nested_list_like,
    is_scalar,
    is_integer,
    is_dict_like,
    is_number,
    is_hashable,
)
from pandas.core.dtypes.common import is_iterator as _is_iterator


def is_list_like_not_nested(obj):
    return is_list_like(obj) and not is_nested_list_like(obj)


def is_dios_like(obj):
    # must have columns
    # columns is some kind of pd.Index
    # iter will iter through columns
    # a `in` obj check if obj is in columns
    # obj[key] will give a pd.Series
    # obj.squeeze() give pd.Series if len(obj) == 1
    return isinstance(obj, DictOfSeries) or isinstance(obj, pd.DataFrame)


def is_iterator(obj):
    """ This is only a dummy wrapper, to warn that the docu of this isnt't right.
    Unlike the example says,
    >>is_iterator([1, 2, 3])
    returns False, not True for lists
    """
    return _is_iterator(obj)


class DictOfSeries:
    """
    DictionaryOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned pd.Series.

    Dataframe vs DictionaryOfSeries
    -------------------------------

    pd.DataFrame :
      - ``df['a']``      indexing with single label, work on columns, return series
      - ``df[iterabel]`` indexing with list, work on columns, return df
      - ``df[slice]``    indexing with slice, work on rows, return df
      - ``df.loc[..]``   loc[cation] based indexing, work on row and/or columns `labels`, return df
      - ``df.iloc[..]``  i[nteger]loc[cation] based indexing, work on row and/or columns, return df

    DictionaryOfSeries:
      - ``dios['a']``       indexing with single label, work on columns, return series
      - ``dios[iterable]``  indexing with list, work on columns, return dios
      - ``dios[slice]``     indexing with slice, work on (all) index'es, return dios
      - ``dios.loc[..]``    loc[cation] based indexing, work on row and/or columns `labels`, return dios
      - ``dios.iloc[..]``   i[nteger]loc[cation] based indexing, work on row and/or columns, return dios

    Todos:
    -----
    todo: to_discuss!! allow any hashable obj as column identifier
         Currently we only allow strings as identifier, to be more df-like we should allow any
         hashable object (unlike df we may should exclude stuff like: ``None`` or ``np.nan`` ??)

    """

    def __init__(self, data=None, columns=None, itype=MixedItype, downcast_policy='save'):

        self._data = pd.Series()
        # We need to keep track of the index-type (itype) of every new Series.
        # If the itypes differ between different series, slicing will almost always fail
        # (eg. a datetime-like slice cannot work on a numeric index and vice versa).
        self._itype = get_itype(itype)

        if downcast_policy not in CAST_POLICIES:
            raise ValueError(f"downcast_policy must be one of {CAST_POLICIES}")
        self._policy = downcast_policy

        if columns is not None and not is_list_like_not_nested(columns):
            raise TypeError("'columns' must be some kind of list-like collection.")

        if data is not None:
            self._init_insert_data(data, columns)

        # insert empty if requested
        if columns is not None:
            for c in columns:
                if c not in self.columns:
                    self._insert(c, pd.Series())

        for s in self._data:
            cast_to_itype(s, self._itype, policy=self._policy, inplace=True)

    def _init_insert_data(self, data, columns):
        def incols(c):
            return c in columns if columns is not None else True

        if isinstance(data, dict) or is_dios_like(data):
            for k in data:
                if incols(k):
                    self._insert(k, data[k])

        elif is_list_like(data):  # also Series !
            data = data if is_nested_list_like(data) else [data]
            if columns is None:
                for i, d in enumerate(data):
                    self._insert(i, d)
            else:
                if len(data) != len(columns):
                    raise ValueError(f"length of passed values is {len(data)}, columns imply {len(columns)}")
                for i, c in enumerate(columns):
                    self._insert(c, data[i])
        else:
            raise ValueError(f"Cannot insert data of type {type(data)}")

    @property
    def columns(self):
        return self._data.index

    @columns.setter
    def columns(self, newindex):
        self._data.index = newindex

    @property
    def values(self):
        return np.array([c.values for c in self._data])

    @property
    def data(self):
        return self._data.values

    @property
    def itype(self):
        return self._itype

    @itype.setter
    def itype(self, newitype):
        itype = get_itype(newitype)

        if not itype_le(self._itype, itype):
            self.__cast_all(itype)

        self._itype = itype

        if not itype.unique:
            throw_MixedItype_err_or_warn(f"Using a {itype} as dios.itype is experimental. As soon as series with "
                                         f"different index types are inserted, slicing will almost always fail. "
                                         f"You are hereby warned!")

    def __cast_all(self, itype):
        k = '?'
        try:
            for k in self._data:
                cast_to_itype(k, itype, policy=self._policy, inplace=True)
        except Exception as e:
            raise type(e)(f"Column {k}: " + str(e)) from e

    def __getitem__(self, key):
        """ dios[?] -> dios/series

        Examples:
          - ``dios['a']``      -> return a pd.Series
          - ``dios[iterable]`` -> return a DictOfSeries of all given columns [1]
          - ``dios[slice]``    -> return a sliced DictOfSeries of all(!) columns in the dios
        Notes:
          - [1] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """
        if is_hashable(key) and key is not None:
            if key not in self.columns:
                # bug-fix(?): integer defaults to iloc ??
                # `s = series([5,6], index=['a','b'])`
                # s[1] returns 6 instead of raising a KeyError
                raise KeyError(f"'{key}'")
            new = self._data[key]
        else:
            keys, ixs, ixalign = self._unpack_key(key)
            new = self.copy_empty()
            for i, k in enumerate(keys):
                # fixme: optimasation: new._data.loc[k] = self._get_item(k, ixs[i], ixalign)
                new[k] = self._get_item(k, ixs[i], ixalign)
        return new

    def _get_item(self, key, ix, ixalign):
        ser = self._data[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        return ser[ix]

    def __setitem__(self, key, value):
        """ dios[x] = y

        Examples:
          - ``dios['a'] = pd.Series()`` -> Set a new pd.Series to an existing column or add it as `new column`.
          - ``dios['a'] = Any`` -> Pass ``any`` to the pd.Series in the corresponding existing(!) column. [1],[2]
          - ``dios[iterable] = Any`` -> Pass ``any`` to the pd.Series's in the corresponding set of columns. [1],[2],[3]
          - ``dios[slice] = Any`` -> Pass ``any`` to all(!) sliced pd.Series's from the current dios. [1],[2]

        Notes:
          - [1] The length of ``Any`` must be be equal to the length of every sliced pd.Series.
          - [2] If ``Any`` is a ``DictOfSeries`` the behavior depends on the option ``dios_to_dios_method``
            in the ``options`` dictionary.
          - [3] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """
        ixalign = False

        if is_hashable(key) and key is not None:
            if key not in self.columns:
                self._insert(key, value)
                return
            else:
                keys, ixs = [key], [slice(None)]
        else:
            keys, ixs, ixalign = self._unpack_key(key)

        assert len(keys) == len(ixs)

        gen = self._unpack_value(keys, ixs, value)
        for tup in gen:
            self._set_item(*tup, ixalign)

    def _set_item(self, key, ix, right, ixalign=False):
        """Set a value (scalar or list or series)"""
        # right can be only a list, a pd.Series or a scalar
        ser = self._data[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        if isinstance(right, pd.Series):
            left = ser[ix]
            right = align_index_by_policy(left, right)
            ix = right.index
        ser.loc[ix] = right

    def _insert(self, col, val):
        """Insert a fresh new value into self"""
        if is_dios_like(val):
            val = val.squeeze()
        elif is_list_like(val) and not is_nested_list_like(val):
            val = pd.Series(val)

        if not isinstance(val, pd.Series):
            raise ValueError(f"Cannot insert data of type {type(val)}")

        val = cast_to_itype(val, self._itype, policy=self._policy)
        self._data.at[col] = val.copy(deep=True)

    def _unpack_value(self, keys, ixs, val):
        """Return a generator that yield (key, indexer, value) for all keys"""
        val = list(val) if is_iterator(val) else val

        if is_dios_like(val):
            val = val.squeeze()

        dioslike, nlistlike = is_dios_like(val), is_nested_list_like(val)

        if (dioslike or nlistlike) and len(val) != len(keys):
            raise ValueError(f"could not broadcast input array with length {len(val)}"
                             f" into dios of length {len(keys)}")

        # now we have everything we need: key, indexer, value
        # so we just pack it nice and cosy and let the caller
        # do the dirty work.
        for i, _ in enumerate(keys):
            key, ix = keys[i], ixs[i]
            if dioslike:
                # we explicitly do not align keys here. usr can use .loc for this
                # purpose, (but we do align on rows, later in the setting chain)
                yield key, ix, val[val.columns[i]]
            elif nlistlike:
                yield key, ix, val[i]
            else:
                yield key, ix, val

    def _unpack_key(self, key):
        """ Determine keys and indexer by type of key. This does not deal
            with single (hashable) label-access, only higher dimension objects
            are handled..
        """
        ixalign = False
        key = list(key) if is_iterator(key) else key

        # bool indexer
        # ------------
        # bool indexer always work on rows, so they need to have
        # an index wo which we can align to. This is necessary
        # because we can hold series of different indices.
        if is_bool_indexer(key):
            if not isinstance(key, pd.Series):
                raise ValueError("Must pass Series with boolean values only")
            keys = self.columns
            indexer, ixalign = [key[key]] * len(keys), True

        elif is_dios_like(key):
            keys = self.columns.intersection(key.columns).to_list()
            for k in keys:
                if not is_bool_indexer(key[k]):
                    raise ValueError("Must pass DictOfSeries with boolean values only")
            indexer, ixalign = [(key[k])[key[k]] for k in keys], True
        # slice
        # -----
        # slices always work rows too, but never fail and
        # doesnt need alignment
        elif isinstance(key, slice):
            keys = self.columns
            indexer = [key] * len(keys)

        # list-like
        # ---------
        # list like stuff, just selects columns
        elif is_list_like(key) and not is_nested_list_like(key):
            key = key.values if isinstance(key, pd.Series) else key
            keys = [k for k in key if k in self.columns]
            fail = [k for k in key if k not in self.columns]
            if fail:
                raise KeyError(f"{fail} not in index")
            indexer = [slice(None)] * len(keys)
        else:
            raise KeyError(f"{type(key)}")
        return keys, indexer, ixalign

    @property
    def loc(self):
        return _LocIndexer(self)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        pd_max_rows = pd.get_option('display.max_rows')
        pd.set_option('display.max_rows', dios_options[OptsFields.disp_max_rows])

        def item2str(k):
            kstr = str(self[k]).replace('\n', '\n ')
            return f'{k}:\n {kstr}\n'

        maxrows = dios_options[OptsFields.disp_max_vars]
        s = ''
        head = maxrows // 2
        tail = len(self.columns) - head
        for i, k in enumerate(self.columns):
            if i < head:
                s += item2str(k)
            elif i == head:
                s += ' ... \t\t\t\t ...\n' * 2
            elif i > tail:
                s += item2str(k)
        pd.set_option('display.max_rows', pd_max_rows)
        return s

    @property
    def empty(self):
        return len(self) == 0 or all(c.empty for c in self._data)

    def all(self):
        return pd.Series([c.all() for c in self._data])

    def any(self):
        return pd.Series([c.any() for c in self._data])

    def aaall(self):
        """absolute all from all"""
        return self.all().all()

    def aaany(self):
        """absolute any of any"""
        return self.any().any()

    def __len__(self):
        return len(self.columns)

    def __iter__(self):
        yield from self.columns.copy()

    def __reversed__(self):
        yield from reversed(self.columns.copy())

    def __contains__(self, item):
        return item in self.columns.copy()

    def __delitem__(self, key):
        del self._data[key]

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        new = self.copy_empty()
        # We use `_data` here, because all checks are already done.
        # So this should be much faster, especially, because we use the underlying dict for
        # getting and setting the values, instead of ``__setitem__`` and ``__getitem__``.
        for i in self._data.index:
            new._data.loc[i] = self._data[i].copy(deep=deep)
        return new

    def copy_empty(self):
        new = DictOfSeries()
        new._itype = self.itype
        return new

    def _op1(self, op):
        new = self.copy_empty()
        try:
            for k in self:
                new[k] = op(self._data[k])
        except Exception as e:
            raise type(e)(f"'{OP_MAP[op]} dios' failed: " + str(e)) from e
        return new

    def _op2(self, op, other, align=True, inplace=False):
        def raiseif(cond, s='lenght'):
            if cond:
                raise ValueError(f"{s} does not match, {s} left: {len(self)}, {s} right: {len(other)} keys")

        def doalign(left, right):
            return left.align(right, join='inner') if align else (left, right)

        def gen():
            if is_dios_like(other):
                raiseif(set(other) != set(self), '#keys')
                for k in self.columns:
                    left, right = self._data[k], other[k]
                    yield k, op(*doalign(left, right))
            elif isinstance(other, pd.Series):
                for k in self.columns:
                    left, right = self._data[k], other
                    yield k, op(*doalign(left, right))
            elif is_dict_like(other):
                raiseif(set(other) != set(self), '#keys')
                for k in self.columns:
                    yield k, op(self._data[k], other[k])
            elif is_nested_list_like(other):
                raiseif(len(other) != len(self), 'length')
                for i, k in enumerate(self.columns):
                    yield k, op(self._data[k], other[i])
            elif is_scalar(other) or is_list_like(other):
                for k in self.columns:
                    yield k, op(self._data[k], other)
            else:
                raise NotImplementedError

        new = self if inplace else self.copy_empty()
        try:
            for k, val in gen():
                new[k] = val
        except Exception as e:
            raise type(e)(f"'dios {OP_MAP[op]} other' failed: " + str(e)) from e
        return new

    __neg__ = partialmethod(_op1, op.neg)
    __abs__ = partialmethod(_op1, op.abs)
    __invert__ = partialmethod(_op1, op.inv)
    __eq__ = partialmethod(_op2, op.eq, align=False)
    __ne__ = partialmethod(_op2, op.ne, align=False)
    __le__ = partialmethod(_op2, op.le, align=False)
    __ge__ = partialmethod(_op2, op.ge, align=False)
    __lt__ = partialmethod(_op2, op.lt, align=False)
    __gt__ = partialmethod(_op2, op.gt, align=False)
    __add__ = partialmethod(_op2, op.add)
    __sub__ = partialmethod(_op2, op.sub)
    __mul__ = partialmethod(_op2, op.mul)
    __mod__ = partialmethod(_op2, op.mod)
    __truediv__ = partialmethod(_op2, op.truediv)
    __floordiv__ = partialmethod(_op2, op.floordiv)
    __pow__ = partialmethod(_op2, op.pow)
    __and__ = partialmethod(_op2, op.and_)
    __or__ = partialmethod(_op2, op.or_)
    __xor__ = partialmethod(_op2, op.xor)

    def squeeze(self):
        return self[self.columns[0]] if len(self) == 1 else self

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    def foreach(self, f, *args, **kwargs):
        self.pipe(f, *args, **kwargs)

    def pipe(self, f, *args, **kwargs):
        """
        Call a pd.Series function on all series in the DictOfSeries

        :param f: pd.Series function to apply on all Series
        :param args: args for the function
        :param kwargs: kwargs for the function
        :return:
        """
        news = pd.Series()
        newd = self.copy_empty()
        need_dios = False
        # return a copy nevertheless, but also run inplace if inplace=True or
        # if the function not has this option, but work inplace.

        inplace = kwargs.get('inplace', False) or kwargs.get('copy', False)

        def getobj(k, inplace):
            if inplace:
                # self[k] may return a copy, so we need to get
                # the item from the dict, by our mother class.
                return self._data[k]
            else:
                return self._data[k].copy()

        for k in self.columns:
            v = f(getobj(k, inplace), *args, **kwargs)
            if inplace:
                continue
            if isinstance(v, pd.Series):
                newd[k] = v
                need_dios = True
            else:
                newd[k] = pd.Series(data=v)
                news[k] = v

        if inplace:
            return None
        if need_dios:
            return newd.squeeze()
        if len(news) == 0:
            return None
        return news.squeeze()


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data
        self._unpack_value = _dios._unpack_value


class _LocIndexer(_Indexer):

    def __init__(self, _dios):
        super().__init__(_dios)
        self._set_item = _dios._set_item

    def __getitem__(self, key):
        rkey, cols, lowdim = self._unpack_key(key)
        if is_scalar(rkey[0]):
            return self._series(rkey, cols, lowdim)
        elif lowdim:
            return self._scalar(rkey[0], cols[0])
        else:
            new = self._dios.copy_empty()
            for i, _ in enumerate(cols):
                c, r = cols[i], rkey[i]
                new[c] = self._data[c].loc[r]
            return new

    def _series(self, rkey, cols, lowdim):
        if lowdim:
            return self._scalar(rkey[0], cols[0])
        new = pd.Series()
        for c in cols:
            try:
                new[c] = self._data[c].loc[rkey]
            except KeyError:
                new[c] = np.nan

    def _scalar(self, r, c):
        return self._data[c].loc[r]

    def __setitem__(self, key, value):
        ixs, keys, _ = self._unpack_key(key)
        gen = self._unpack_value(keys, ixs, value)
        for tup in gen:
            self._set_item(*tup)

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        lowdim = False
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            ckey = list(ckey) if is_iterator(ckey) else ckey

            # determine columns
            if is_nested_list_like(ckey) or is_dios_like(ckey):
                raise ValueError("Cannot index with multidimensional key")
            if isinstance(ckey, str):
                cols = [ckey]
                lowdim = True
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            else:
                try:
                    # list, boolean-list or series
                    cols, *_ = self._dios._unpack_key(ckey)
                except Exception as e:
                    raise e
        else:
            cols = list(self._data.index)
            rkey = key
        # blowup
        rkey = [rkey] * len(cols)
        return rkey, cols, lowdim

    def _col_slice_to_col_list(self, cslice):
        """ see here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-slicing-with-labels
        """
        keys = list(self._data.index)
        try:
            start = keys.index(cslice.start) if cslice.start is not None else None
            stop = keys.index(cslice.stop) if cslice.stop is not None else None
        except ValueError:
            raise KeyError("The slice start label, or the slice stop label, is not present in columns.")
        if not is_integer(cslice.step) or cslice.step <= 0:
            return []
        return keys[slice(start, stop + 1, cslice.step)]


class _iLocIndexer(_Indexer):

    def __getitem__(self, key):
        rkey, cols, lowdim = self._unpack_key(key)
        if is_scalar(rkey[0]):
            return self._series(rkey, cols, lowdim)
        elif lowdim:
            return self._scalar(rkey[0], cols[0])
        else:
            new = self._dios.copy_empty()
            for i, _ in enumerate(cols):
                c, r = cols[i], rkey[i]
                new[c] = self._data[c].iloc[r]
            return new

    def _series(self, rkey, cols, lowdim):
        if lowdim:
            return self._scalar(rkey[0], cols[0])
        new = pd.Series()
        for c in cols:
            try:
                new[c] = self._data[c].iloc[rkey]
            except KeyError:
                new[c] = np.nan

    def _scalar(self, r, c):
        return self._data[c].iloc[r]

    def __setitem__(self, key, value):
        ixs, keys, _ = self._unpack_key(key)
        gen = self._unpack_value(keys, ixs, value)
        for tup in gen:
            self._set_item_positional(*tup)
        raise NotImplemented

    def _set_item_positional(self, key, ix, val):
        ser = self._data[key]
        if isinstance(val, pd.Series):
            index = ser.iloc[ix].index
            index = index.intersection(val.index)
            if not index.empty:
                ser.loc[index] = val.loc[index].copy()
        else:
            ser.iloc[ix] = val

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        lowdim = False
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            ckey = list(ckey) if is_iterator(ckey) else ckey

            # determine columns
            if is_integer(ckey):
                self._check_keys([ckey])
                cols = self._integers_to_col_list([ckey])
                lowdim = True
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey) and not is_nested_list_like(ckey):
                arr = np.array(ckey)
                if is_bool_array(arr):
                    raise NotImplementedError
                self._check_keys(ckey)
                cols = self._integers_to_col_list(ckey)
            elif isinstance(ckey, pd.Series):
                raise NotImplementedError
            elif is_bool_indexer(ckey):
                raise NotImplementedError
            else:
                raise KeyError(f"{ckey} of type {type(ckey)}")
        else:
            cols = list(self._data.index)
            rkey = key

        # blowup
        rkey = [rkey] * len(cols)
        return rkey, cols, lowdim

    def _check_keys(self, keys):
        bound = len(self._data)
        for k in keys:
            if not is_integer(k):
                raise ValueError(f"{type(k)} is not integer")
            if k not in range(-bound, bound):
                raise KeyError("positional indexer(s) are out-of-bounds in columns")

    def _integers_to_col_list(self, ints):
        klist = list(self._data.index)
        ks = set()
        for i in ints:
            ks.add(klist[i])
        return list(ks)

    def _col_slice_to_col_list(self, sl):
        for s in [sl.start, sl.stop, sl.step]:
            if not is_integer(s):
                raise TypeError(f"positional indexing with slice must be integers, passed type was {type(s)}")
        return list(self._data.index)[sl]
